/*
 * (c) Copyright 2001 MyCorporation.
 * All Rights Reserved.
 */
package org.eclipse.jdt.internal.debug.eval;

import java.util.ArrayList;
import java.util.Iterator;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.core.IBuffer;
import org.eclipse.jdt.core.IBufferFactory;
import org.eclipse.jdt.core.IClassFile;
import org.eclipse.jdt.core.ICodeCompletionRequestor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.ICompletionRequestor;
import org.eclipse.jdt.core.IImportContainer;
import org.eclipse.jdt.core.IImportDeclaration;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaModel;
import org.eclipse.jdt.core.IJavaModelStatusConstants;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageDeclaration;
import org.eclipse.jdt.core.ISourceRange;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.IWorkingCopy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Message;
import org.eclipse.jdt.debug.core.IJavaClassType;
import org.eclipse.jdt.debug.core.IJavaDebugTarget;
import org.eclipse.jdt.debug.core.IJavaObject;
import org.eclipse.jdt.debug.core.IJavaStackFrame;
import org.eclipse.jdt.debug.core.IJavaThread;
import org.eclipse.jdt.debug.core.IJavaValue;
import org.eclipse.jdt.debug.eval.IEvaluationEngine;
import org.eclipse.jdt.debug.eval.IEvaluationListener;
import org.eclipse.jdt.internal.core.JavaModelStatus;
//import org.eclipse.jdt.internal.core.newbuilder.NameEnvironment;
import org.eclipse.jdt.internal.debug.core.model.JDIDebugTarget;
import org.eclipse.jdt.internal.eval.EvaluationConstants;
import org.eclipse.jdt.internal.core.eval.ASTAPIVisitor;
import org.eclipse.jdt.core.eval.model.IClassType;
import org.eclipse.jdt.core.eval.model.IValue;
import org.eclipse.jdt.core.eval.model.IVariable;


/**
 * @version 	1.0
 * @author
 */
public class ASTAPIEvaluationEngine implements IEvaluationEngine {

	private IJavaProject fProject;
	
	private IJavaDebugTarget fDebugTarget;

	public ASTAPIEvaluationEngine(IJavaProject project, IJavaDebugTarget debugTarget) {
		setJavaProject(project);
		setDebugTarget(debugTarget);
	}
	
	public void setJavaProject(IJavaProject project) {
		fProject = project;
	}

	public void setDebugTarget(IJavaDebugTarget debugTarget) {
		fDebugTarget = debugTarget;
	}

	/*
	 * @see IEvaluationEngine#evaluate(String, IJavaThread, IEvaluationListener)
	 */
	public void evaluate(
		String snippet,
		IJavaThread thread,
		IEvaluationListener listener)
		throws DebugException {
	}

	/*
	 * @see IEvaluationEngine#evaluate(String, IJavaStackFrame, IEvaluationListener)
	 */
	public void evaluate(
		String snippet,
		IJavaStackFrame frame,
		IEvaluationListener listener)
		throws DebugException {
			IJavaProject javaProject = getJavaProject();
			RuntimeContext context = new RuntimeContext(javaProject, frame);

			ASTCodeSnippetToCuMapper mapper = null;
			CompilationUnit unit = null;
			EvaluationResult result = new EvaluationResult(this, snippet, (IJavaThread)frame.getThread());

			try {
				
				IVariable[] locals = context.getLocals();
				int numLocals= locals.length;
				int[] localModifiers = new int[locals.length];
				String[] localTypesNames= new String[numLocals];
				String[] localVariables= new String[numLocals];
				for (int i = 0; i < numLocals; i++) {
					localVariables[i] = locals[i].getName();
					localTypesNames[i] = ((EvaluationValue)locals[i].getValue()).getJavaValue().getReferenceTypeName();
					localModifiers[i]= 0;
				}
				
				mapper = new ASTCodeSnippetToCuMapper(new String[0], frame.getDeclaringType().getName(), localModifiers, localTypesNames, localVariables, frame, snippet);

				unit = AST.parseCompilationUnit(mapper.getSource().toCharArray(), mapper.getCompilationUnitName(), javaProject);
			} catch (JavaModelException e) {
				throw new DebugException(e.getStatus());
			} catch (CoreException e) {
				throw new DebugException(e.getStatus());
			}
			
			Message[] messages= unit.getMessages();
			if (messages.length != 0) {
				boolean error= false;
				StringBuffer buffer= new StringBuffer();
				for (int i = 0; i < messages.length; i++) {
					String message= messages[i].getMessage();
					// bad test for remove the "Void methods cannot return a value" error
					if (!"Void methods cannot return a value".equals(messages[i].getMessage())) {
						buffer.append(messages[i].getSourcePosition() + " : " + messages[i].getMessage() + "\n");
						error = true;
					}
				}
				if (error) {
					result.setValue(((EvaluationValue)context.getVM().newValue(buffer.toString())).getJavaValue());
					listener.evaluationComplete(result);
					return;
				}
			}
			
			ASTAPIVisitor visitor = new ASTAPIVisitor(mapper.getStartPosition());
			unit.accept(visitor);

			if (visitor.hasErrors()) {
				StringBuffer buffer = new StringBuffer();
				for (Iterator iterator = visitor.getErrors().iterator(); iterator.hasNext();) {
					buffer.append(((ASTAPIVisitor.ErrorMessage) iterator.next()).getMessage() + "\n");
				}
				result.setValue(((EvaluationValue)context.getVM().newValue(buffer.toString())).getJavaValue());
				listener.evaluationComplete(result);
				return;
			}
			
			IValue value = null;
			try {
				value = visitor.getInstructions().evaluate(context);
			} catch (CoreException e) {
				throw new DebugException(e.getStatus());
			}
			
			IJavaValue jv = ((EvaluationValue)value).getJavaValue();
			result.setValue(jv);
			listener.evaluationComplete(result);
	}


	/*
	 * @see IEvaluationEngine#evaluate(String, IJavaObject, IJavaThread, IEvaluationListener)
	 */
	public void evaluate(
		String snippet,
		IJavaObject thisContext,
		IJavaThread thread,
		IEvaluationListener listener)
		throws DebugException {
	}

	/*
	 * @see IEvaluationEngine#getJavaProject()
	 */
	public IJavaProject getJavaProject() {
		return fProject;
	}

	/*
	 * @see IEvaluationEngine#getDebugTarget()
	 */
	public IJavaDebugTarget getDebugTarget() {
		return fDebugTarget;
	}

	/*
	 * @see IEvaluationEngine#dispose()
	 */
	public void dispose() {
	}
	

}
